#* Functions for calculating the power flow in a power grid described by a NDD
#*------------------------------------------------------------------------------

#=
Adds or updates the power flow solution in a NDD using a power flow solution obtained with PowerModels.jl and returns the NDD. If model=:ac, the AC power flow equations are used to calculate the flows in the branches. If model=:dc, the DC power flow equations are used. 
=#
function update_pf_data!(
        network_data::Dict{String,<:Any}, 
        pf_result::Dict{String,<:Any};
        model = :ac # how to calculate power flows
    )
    
    ### Add solution (voltage angles and magnitudes) to NDD
    update_data!(network_data, pf_result["solution"])
    
    ### Calculate power flows with PowerModels.jl depending on chosen model
    if model == :dc
        flows = calc_branch_flow_dc(network_data) # DC-PF
        update_data!(network_data, flows) # add power flows to NDD
        calc_branchloads!(network_data, model=:dc)
    elseif model == :ac
        flows = calc_branch_flow_ac(network_data) # AC-PF
        update_data!(network_data, flows) # add power flows to NDD
        calc_branchloads!(network_data, model=:ac)
    else
        throw(ArgumentError("Unknown model $model."))
    end

    return network_data
end

#*------------------------------------------------------------------------------

#=
Reads the power grid data saved in a file and calculates the AC (model=:ac) or DC (model=:dc) power flow solution using PowerModels.jl. The solution is added to a network data dictionary (NDD) of PowerModels.jl that is then returned.
=#
function calc_init_op(file::String; model=:ac)
    network_data = parse_file(file, import_all=false) # read data file
    simplify_network!(network_data)
    
    ### Calculate power flow for chosen model
    if model == :dc
        calc_dc_pf!(network_data) # calculate and add DC-PF
    elseif model == :ac
        calc_ac_pf!(network_data) # calculate and add AC-PF
    else
        throw(ArgumentError("Unknown model $model."))
    end

    return network_data # return NDD
end
##

#=
Calculates the AC power flow solution for the given NDD and updates it in the NDD.
CAUTION: Overwrites any previous AC solution contained in the NDD!
=#
function calc_ac_pf!(
        network_data::Dict{String,<:Any}, 
        method = :JuMP; 
        print_level = 5
    )

    if method == :JuMP # calculate AC-PF solution using a JuMP model
        pf_result = run_ac_pf(
            network_data, 
            optimizer_with_attributes(
                Ipopt.Optimizer, 
                "print_level" => print_level
            )
        )
    elseif method == :NLsolve # calculate AC-PF solution using NLsolve.jl
        pf_result = Dict{String,Any}()
        pf_result["solution"] = compute_ac_pf(network_data, show_trace=true)
    else
        throw(ArgumentError("Unknown method $method."))
    end

    update_pf_data!(network_data, pf_result, model=:ac) # update PF in NDD
    return nothing
end

#=
Calculates the DC power flow solution for the given NDD and updates it in the NDD.
CAUTION: Overwrites any previous DC solution contained in the NDD!
=#
function calc_dc_pf!(network_data::Dict{String,<:Any}; print_level = 5)
    ### Calculate AC-PF solution using PowerModels.jl and Ipopt
    pf_result = run_dc_pf(
        network_data, 
        optimizer_with_attributes(
            Ipopt.Optimizer, 
            "print_level" => print_level
        )
    )

    update_pf_data!(network_data, pf_result, model=:dc) # update PF in NDD
    return nothing
end

#=
Calculates the loading (flow/capacity) of all branches in the NDD based on the current power flow solution and adds it to the NDD. If model=:dc, only the loading due to the active power flow is calculated. For model=:ac, also the loading due to reactive power and apparent power is calculated.
=#
function calc_branchloads!(network_data::Dict{String,<:Any}; model=:ac)
    branchloads = Dict{String,Any}(
        "baseMVA"  => 100.0,
        "branch" => Dict{String,Any}(), 
        "per_unit" => true
    )
    L = length(network_data["branch"]) # number of branches
    sizehint!(branchloads["branch"], L)

    ### Go through branches and calculate power flows
    for (i, branch) in network_data["branch"]
        if model == :dc # DC-PF
            ### Active power flow
            mw_fr = sqrt(branch["pf"]^2)
            mw_to = sqrt(branch["pt"]^2)
            ### Add active power flow to branchloads
            branchloads["branch"][i] = Dict(
                "MW-loading" => max(mw_fr, mw_to) / branch["rate_a"]
            )
        elseif model == :ac # AC-PF
            ### Active power flow
            mw_fr = sqrt(branch["pf"]^2)
            mw_to = sqrt(branch["pt"]^2)
            ### Reactive power flow
            mvar_fr = sqrt(branch["qf"]^2)
            mvar_to = sqrt(branch["qt"]^2)
            ### Apparent power flow
            mva_fr = sqrt(branch["pf"]^2 + branch["qf"]^2)
            mva_to = sqrt(branch["pt"]^2 + branch["qt"]^2)
            ### Add all flows to branchloads
            branchloads["branch"][i] = Dict(
                "MW-loading" => max(mw_fr, mw_to) / branch["rate_a"],
                "Mvar-loading" => max(mvar_fr, mvar_to) / branch["rate_a"],
                "MVA-loading" => max(mva_fr, mva_to) / branch["rate_a"]
            )
        else
            throw(ArgumentError("Unknown model $model."))
        end
    end

    update_data!(network_data, branchloads) # add branchloads to NDD
    return nothing
end

#*------------------------------------------------------------------------------

#=
Calculates the total impact of a hurricane given a time series of damaged transmission lines. Each time a transmission line is destroyed, the AC power flow equations are solved and failure cascades may unfold. The terminal output of the solver is saved in a .txt file, if an "outputpath" is defined. Additionally the final network data dictionary is saved in a .jld2 file, if a "savepath" is defined.
=#
function calc_total_impact!(
        network_data::Dict{String,<:Any},
        tl_damage::Array{Tuple{String,Int64},1}; # time series of tl damage
        method = :JuMP, # solver to use
        print_level = 5, # Ipopt output details
        outputpath = "", # path for saving solver terminal output
        savepath = "", # path for saving final network data dictionary
    )

    if isempty(outputpath) == true # calculate cascade without saving output
        _calc_total_impact!(
            network_data, tl_damage, method=method, print_level=print_level
        )
    else # calculate cascade and save output
        output = @capture_out _calc_total_impact!(
            network_data, tl_damage, method=method, print_level=print_level
        )
        ### Write terminal output of solver into file
        open(outputpath, "w") do io
            write(io, output)
        end
    end

    if isempty(savepath) == false # save final network data
        save(savepath, "network_data", network_data)
    end

    return nothing
end

#=
See calc_total_impact!
=#
function _calc_total_impact!(
        network_data::Dict{String,<:Any},
        tl_damage::Array{Tuple{String,Int64},1}; # time series of tl damage
        method = :JuMP, # solver to use
        print_level = 0 # Ipopt output
    )

    ### Get unique frames in which transmission lines were damaged
    unique_frames = sort(unique([tl_damage[i][2] for i in 1:length(tl_damage)]))

    ### Go through frames in which transmission lines were destroyed
    for f in unique_frames
        destroyed_tl = [dmg[1] for dmg in tl_damage if dmg[2] == f]
        destroy_tl!(network_data, destroyed_tl)

        ### Resolve AC-PF and see whether a cascade unfolds
        calc_cascade!(network_data, method, print_level)
    end
    
    return nothing
end

#=
Calculates whether a cascading failure unfolds in a damaged network. If any network contained in network_data misses a slack bus, the largest generator is selected in each connected component. Additionally, the active power balance is restored for each connected component in every step, if it is violated.
=#
function calc_cascade!(
        network_data::Dict{String,<:Any},
        method = :JuMP,
        print_level = 0
    )

    ### Update network topology (remove dangling buses etc.)
    # simplify_network!(network_data)

    ### Define new slacks, if necessary
    correct_reference_buses!(network_data)
    
    ### Restore active power balance
    restore_p_balance!(network_data)

    ### Calculate new AC-PF solution
    set_ac_pf_start_values!(network_data) # last state as initial conditions
    calc_ac_pf!(network_data, method, print_level=print_level)
    
    ### Check for overloaded branches
    overloaded_br = get_branches(network_data, min_loading=1.)

    if isempty(overloaded_br) == false
        ### Print overloaded branches as output
        println("-------------------------------------------------------------")
        println("Overloaded branches:", overloaded_br)
        println("-------------------------------------------------------------")
        
        ### Deactivate overloaded branches
        disable_branches!(network_data, [br[1] for br in overloaded_br])
        ### Repeat cascade algorithm
        calc_cascade!(network_data, method, print_level)
    end

    return nothing
end

#=
Identifies connected components in the network and restores active power balance in each component.
=#
function restore_p_balance!(network_data::Dict{String,<:Any})
    ### Calculate connected components
    connected_components = calc_connected_components(network_data)

    ### Restore active power balance for every connected component
    for cc in connected_components
        _restore_p_balance!(network_data, cc)
    end

    ### Update network topology (remove dangling buses etc.)
    simplify_network!(network_data)

    return nothing
end

#=
Restores active power balance in a connected component identified by the set of buses cc. Depending on whether an overproduction or underproduction exists, generator dispatches may be increased or reduced. If necessary, load can be shed by reducing the demands uniformly until the total demand matches generation. For more details see flow chart. 
=#
function _restore_p_balance!(network_data::Dict{String,<:Any}, cc::Set{Int64})
    active_cc_gens = Int64[
        gen["index"] for gen in values(network_data["gen"]) if 
        gen["gen_bus"] ∈ cc && gen["gen_status"] != 0
    ] # indices of active generators that are connected to buses in cc

    active_cc_loads = Int64[
        load["index"] for load in values(network_data["load"]) if 
        load["load_bus"] ∈ cc && load["status"] != 0
    ] # indices of active loads that are connected to buses in cc

    ### Calculate active power mismatch
    cc_Δp, cc_Δp_lim = calc_Δp(
        network_data, active_cc_gens, active_cc_loads, cc
    )

    if isapprox(cc_Δp, 0, atol=1e-10) # "sufficient" power balance
        nothing # nothing has to be done
    elseif cc_Δp > 0 # overproduction
        if isapprox(cc_Δp_lim, 0, atol=1e-10) # power balance at minimum
            ### Set all active generators to their minimum
            for g in active_cc_gens
                id = string(g)
                network_data["gen"][id]["pg"] = network_data["gen"][id]["pmin"]
            end
        elseif cc_Δp_lim < 0 # sufficient dispatch reduction possible
            ### Reduce all dispatches uniformly
            _reduce_p_dispatch!(
                network_data, active_cc_gens, active_cc_loads, cc, cc_Δp
            )
        else # overproduction is unavoidable
            ### Fail complete component
            _deactivate_cc!(network_data, active_cc_gens, active_cc_loads)
        end
    elseif cc_Δp < 0 # underproduction
        if isapprox(cc_Δp_lim, 0, atol=1e-10) # power balance at maximum
            ### Set all active generators to their maximum
            for g in active_cc_gens
                id = string(g)
                network_data["gen"][id]["pg"] = network_data["gen"][id]["pmax"]
            end
        elseif cc_Δp_lim < 0 # total demand cannot be matched
            ### Set all active generators to their maximum
            for g in active_cc_gens
                id = string(g)
                network_data["gen"][id]["pg"] = network_data["gen"][id]["pmax"]
            end
            ### Shed the minimum amount of load
            _shed_minimum_load!(
                network_data, active_cc_gens, active_cc_loads, cc, cc_Δp_lim
            )
        else # sufficient dispatch increase possible
            ### Increase all dispatches uniformly
            _increase_p_dispatch!(
                network_data, active_cc_gens, active_cc_loads, cc, cc_Δp
            )
        end
    end

    return nothing
end

#=
Calculates and returns the active power mismatch in a connected component identified by the set of buses cc. Depending on whether an overproduction or underproduction exists, the limiting mismatch with minimum or maximum possible active power generation is also calculated and returned.
=#
function calc_Δp(
        network_data::Dict{String,<:Any},
        active_cc_gens::Array{Int64,1}, 
        active_cc_loads::Array{Int64,1},
        cc::Set{Int64}
    )

    ### Calculate active power demand in connected component
    cc_pd = sum(
        Float64[network_data["load"]["$l"]["pd"] for l in active_cc_loads]
    )
    
    ### Calculate total active power loss in branches in connected component
    cc_ploss = sum(
        Float64[br["pt"] + br["pf"] for br in values(network_data["branch"]) 
        if br["f_bus"] ∈ cc && br["t_bus"] in cc && br["br_status"] == 1]
    )

    ### Calculate active power generation in connected component
    cc_pg = sum(
        Float64[network_data["gen"]["$g"]["pg"] for g in active_cc_gens]
    )

    ### Calculate relevant mismatches
    cc_Δp = cc_pg - cc_pd - cc_ploss 
    if isapprox(cc_Δp, 0, atol=1e-10) # "sufficient" power balance
        cc_Δp = 0.0
        cc_Δp_lim = NaN # limiting mismatch not needed
    elseif cc_Δp > 0 # overproduction
        cc_pmin = sum(
            Float64[network_data["gen"]["$g"]["pmin"] for g in active_cc_gens]
        ) # minimum possible active power dispatch
        cc_Δp_lim = cc_pmin - cc_pd - cc_ploss # limiting mismatch
    elseif cc_Δp < 0 # underproduction
        cc_pmax = sum(
            Float64[network_data["gen"]["$g"]["pmax"] for g in active_cc_gens]
        ) # maximum possible active power dispatch
        cc_Δp_lim = cc_pmax - cc_pd - cc_ploss # limiting mismatch
    end

    return cc_Δp, cc_Δp_lim
end

#=
Uniformly reduces active power dispatches of active generators in a connected component according to the active power mismatch. This is done recursively, since generators may hit their minimum dispatch, which has to be compensated by other generators.
=#
function _reduce_p_dispatch!(
        network_data::Dict{String,<:Any},
        active_cc_gens::Array{Int64,1},
        active_cc_loads::Array{Int64,1},
        cc::Set{Int64},
        cc_Δp::Float64 # positive mismatch
    )

    ### Find active generators that are able to reduce their dispatch
    capable_cc_gens = Int64[
        g for g in active_cc_gens if 
        network_data["gen"]["$g"]["pg"] > network_data["gen"]["$g"]["pmin"]
    ]
    g_Δp = cc_Δp / length(capable_cc_gens) # power reduction for each generator

    ### Reduce dispatches as much as possible 
    for g in capable_cc_gens
        gen = network_data["gen"]["$g"]
        if g_Δp >= gen["pg"] - gen["pmin"] # generator reaches minimum dispatch
            gen["pg"] = gen["pmin"]
        else # generator can reduce dispatch sufficiently
            gen["pg"] -= g_Δp
        end
    end

    ### Recalculate active power mismatch and repeat reduction, if necessary
    cc_Δp, cc_Δp_lim = calc_Δp(
        network_data, active_cc_gens, active_cc_loads, cc
    )
    if !isapprox(cc_Δp, 0, atol=1e-10) # power balance still not restored
        _reduce_p_dispatch!(
            network_data, active_cc_gens, active_cc_loads, cc, cc_Δp
        ) # repeat
    end

    return nothing
end

#=
Uniformly increases active power dispatches of active generators in a connected component according the active power mismatch. This is done recursively, since generators may hit their maximum dispatch, which has to be compensated by other generators.
=#
function _increase_p_dispatch!(
        network_data::Dict{String,<:Any},
        active_cc_gens::Array{Int64,1},
        active_cc_loads::Array{Int64,1},
        cc::Set{Int64},
        cc_Δp::Float64 # negative mismatch
    )
    
    ### Find active generators that are able to increase their dispatch
    capable_cc_gens = Int64[
        g for g in active_cc_gens if 
        network_data["gen"]["$g"]["pg"] < network_data["gen"]["$g"]["pmax"]
    ]
    g_Δp = abs(cc_Δp) / length(capable_cc_gens) # power increase per generator

    ### Increase dispatches as much as possible 
    for g in capable_cc_gens
        gen = network_data["gen"]["$g"]
        if g_Δp >= gen["pmax"] - gen["pg"] # generator reaches maximum dispatch
            gen["pg"] = gen["pmax"]
        else # generator can reduce dispatch sufficiently
            gen["pg"] += g_Δp
        end
    end

    ### Recalculate active power mismatch and repeat increase, if necessary
    cc_Δp, cc_Δp_lim = calc_Δp(
        network_data, active_cc_gens, active_cc_loads, cc
    )
    if !isapprox(cc_Δp, 0, atol=1e-10) # power balance still not restored
        _increase_p_dispatch!(
            network_data, active_cc_gens, active_cc_loads, cc, cc_Δp
        ) # repeat
    end

    return nothing
end

#=
Sheds the amount of load necessary to restore active power balance in a connected component. If active generators are already operated at their maximum possible dispatch, the amount of load shed should be optimal (minimal). 
=#
function _shed_minimum_load!(
        network_data::Dict{String,<:Any},
        active_cc_gens::Array{Int64,1},
        active_cc_loads::Array{Int64,1},
        cc::Set{Int64},
        cc_Δp_lim::Float64 # negative limiting mismatch
    )

    ### Find active loads that are able to reduce their demand
    capable_cc_loads = Int64[
        l for l in active_cc_loads if network_data["load"]["$l"]["pd"] > 0
    ]
    l_Δp = abs(cc_Δp_lim) / length(capable_cc_loads) # demand reduction per load

    ### Reduce demands as much as possible 
    for l in capable_cc_loads
        load = network_data["load"]["$l"]
        if l_Δp >= load["pd"] # load reaches zero demand
            load["pd"] = 0.0
        else # load can reduce its demand sufficiently
            load["pd"] -= l_Δp
        end
    end

    ### Recalculate active power mismatch and repeat load shedding, if necessary
    cc_Δp, cc_Δp_lim = calc_Δp(
        network_data, active_cc_gens, active_cc_loads, cc
    )
    if !isapprox(cc_Δp, 0, atol=1e-10) # power balance still not restored
        _shed_minimum_load!(
            network_data, active_cc_gens, active_cc_loads, cc, cc_Δp_lim
        ) # repeat
    end

    return nothing
end
